from __future__ import division

import uuid
import random
import csv
from copy import copy
from datetime import date, timedelta
from enum import Enum


import numpy as np
import pandas as pd
from faker import Faker
from tqdm import tqdm

fake = Faker()
nicknames = pd.read_csv('nicknames.csv', skipinitialspace=True)

class CategoricalDistribution(Enum):
    def __new__(cls, prob):
        # Extend the AutoNumber example to allow dupe values
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value, prob
        return obj
    @classmethod
    def __probabilities__(cls):
        total = sum(m.value[1] for m in cls)
        return [m.value[1]/total for m in cls]
class Sex(CategoricalDistribution):
    male = 50
    female = 50
# Values from http://www.census.gov/prod/cen2010/briefs/c2010br-02.pdf
class Race(CategoricalDistribution):
    white     = 223553265
    black     =  38929319
    amindian  =   2932248
    asian     =  14674252
    pacisland =    540013
    other     =  19107368 + 9009073
class Ethnicity(CategoricalDistribution):
    hispanic    =  50477594
    nonhispanic = 258267944
    
    @classmethod
    def __probabilities_by_race__(cls, race):
        if race == Race.white:
            vals = [29184290, 201856108]
        elif race == Race.black:
            vals = [1897218, 40123525]
        elif race == Race.amindian:
            vals = [1190904, 4029675]
        elif race == Race.asian:
            vals = [598146, 16722710]
        elif race == Race.pacisland:
            vals = [210307, 1014888]
        elif race == Race.other:
            vals = [20714218, 1033866]
        else: error()
        tot = sum(vals)
        return [v/tot for v in vals]

class Person(object):
    def __init__ (self,
            fname,
            lname,
            ssn,
            sex,
            dob,
            race,
            ethnicity):
        self.id = uuid.uuid4()
        self.fname = fname
        self.lname = lname
        self.ssn = ssn
        self.sex = sex
        self.dob = dob
        self.race = race
        self.ethnicity = ethnicity
        self.name_change_count = 0
        
    def write_row(self, csvfile):
        csvfile.writerow([
                self.id,
                self.munged_fname(),
                self.munged_lname(),
                self.munged_ssn(),
                self.munged_sex(),
                self.munged_dob(),
                self.munged_race(),
                self.munged_ethnicity(),
            ])
    
    def munged_fname(self):
        fname = self.fname
        if fname in nicknames.name.values and random.random() < .2:
            fname = random.choice(list(nicknames[nicknames.name == fname].nickname))
        if fname[-1] == 'y' and random.random() < 0.1:
            fname = fname[:-1] + 'ie'
        elif fname[-2:] == 'ie' and random.random() < 0.1:
            fname = fname[:-2] + 'y'
        return typo(fname)
    
    def munged_lname(self):
        if self.sex == Sex.female and self.name_change_count < 2 and random.random() < 0.05:
            # Name change. TODO: this is really a rate per year, not per record
            self.lname = fake.last_name()
            self.name_change_count = self.name_change_count + 1
        lname = self.lname
        if ' ' in self.lname:
            if random.random() < .4:
                lname = lname.split(' ')[0]
            elif random.random() < .2:
                lname = lname.split(' ')[1]
            elif random.random() < .1:
                lname = lname.replace(' ', '-')
        return typo(lname)
    
    def munged_ssn(self):
        if random.random() < .15: return None
        if random.random() < .01:
            chars = list(self.ssn)
            for i, d in enumerate(chars):
                if d == '-': continue
                if random.random() < 1/9:
                    chars[i] = str((int(d) + random.choice([-1,1])) % 10)
            return ''.join(chars)
        return self.ssn
    
    def munged_sex(self):
        if random.random() < .05: return None
        if random.random() < .001: return 'F' if self.sex == Sex.male else 'M'
        return self.sex.name[0].upper()
    
    def munged_dob(self):
        if random.random() < .05: return None
        dob = self.dob
        r = random.random()
        if dob.day <= 12 and r < 0.01:
            return date(dob.year,dob.day,dob.month)
        if dob.month < 12 and r < 0.02:
            return date(dob.year,dob.month+1,min(dob.day, 28))
        if dob.month > 1 and r < 0.03:
            return date(dob.year,dob.month-1,min(dob.day, 28))
        if dob.day < 28 and r < 0.04:
            return date(dob.year,dob.month,dob.day+1)
        if dob.day > 1 and r < 0.05:
            return date(dob.year,dob.month,dob.day-1)
        if dob.day > 10 and r < 0.06:
            return date(dob.year,dob.month,dob.day-10)
        if dob.day < 19 and r < 0.07:
            return date(dob.year,dob.month,dob.day+10)
        if r < 0.09:
            return date(dob.year + random.choice((-1,1)), dob.month, min(dob.day, 28))
        if r < 0.15: # Birthday issues are pretty common
            return dob + timedelta(days=random.normalvariate(0,365/2))
        return dob
    
    def munged_race(self):
        if random.random() < .2: return None
        if random.random() < .1: return random.choice(list(Race)).name
        return self.race.name
    
    def munged_ethnicity(self):
        if random.random() < .3: return None
        if random.random() < .1: return random.choice(list(Ethnicity)).name
        return self.ethnicity.name
    
    def twin(self):
        p = copy(self)
        # A new, different person
        p.id = uuid.uuid4()
        # Random gender and new name
        ismale = fake.boolean()
        while p.fname == self.fname: # ensure we don't give the same name
            p.fname = fake.first_name_male() if ismale else fake.first_name_female()
        p.sex = Sex.male if ismale else Sex.female
        # SSN off by one
        s = str(int(p.ssn.replace('-',''))+1).zfill(9)
        p.ssn = '-'.join((s[:3], s[3:5], s[5:]))
        return p
    
    @staticmethod
    def _rand():
        ismale = fake.boolean()
        fname = fake.first_name_male() if ismale else fake.first_name_female()
        race = np.random.choice(Race, 1, p=Race.__probabilities__())[0]
        ethn = np.random.choice(Ethnicity, 1, p=Ethnicity.__probabilities_by_race__(race))[0]
        # Let's just guess that 50% of hispanics have two last names, and a small number otherwise:
        lname = fake.last_name()
        if (ethn == Ethnicity.hispanic and random.random() < .5) or random.random() < .02:
            lname = lname + " " + fake.last_name()
        return Person(fname, lname, fake.ssn(), Sex.male if ismale else Sex.female,
                      fake.date_time_between_dates(date(1940, 1, 1), date.today()).date(),
                      race, ethn)
    
    @staticmethod
    def rand(n=None):
        if n is None: return Person._rand()
        return [Person._rand() for _ in tqdm(range(n),'creating people')]

def typo(s):
    chars = list(s)
    i = 0
    while i < len(chars):
        if random.random() < .00033333:
            chars.insert(i, chars[i])
            i = i+1
        if random.random() < .00033333 and 0 <= ord(chars[i].lower()) - ord('a') < 26:
            chars[i] = chr(((ord(chars[i].lower()) - ord('a') + random.choice(range(1,25))) % 26) + ord('a'))
        if random.random() < .00033333:
            chars.pop(i)
            i = i-1
        i = i+1
    return ''.join(chars)

def create_population(n=100, twin_rate=.025):
    pop = Person.rand(int(n * (1-twin_rate)))
    twins = np.random.choice(pop, n - len(pop))
    for t in tqdm(twins, 'adding twins'):
        pop.append(t.twin())
    return pop

def create_csv(pop, filename, mean=20):
    with open(filename,'w') as fd:
        csvwriter = csv.writer(fd)
        csvwriter.writerow(['uuid','first_name','last_name','ssn','sex','dob','race','ethnicity'])
        for p in tqdm(pop, desc='writing csv'):
            for _ in range(int(random.expovariate(1/mean))+1):
                p.write_row(csvwriter)

if __name__ == '__main__':
    population = create_population(20000)
    create_csv(population, 'people.csv')



# -*- coding: utf-8 -*-

"""
Based on: https://github.com/datamade/dedupe-examples/tree/master/pgsql_big_dedupe_example
"""
import os
import csv
import tempfile
import time
import logging

import json
import yaml
import psycopg2 as psy
import psycopg2.extras

import click
import dedupe

import exact_matches

START_TIME = time.time()


@click.command()
@click.option('--config', help='YAML- or JSON-formatted configuration file.')
@click.option('--db', help='YAML- or JSON-formatted database connection credentials.')
def main(config, db, verbosity=2):
    log_level = logging.WARNING
    if verbosity == 1:
        log_level = logging.INFO
    elif verbosity is None or verbosity >= 2:
        log_level = logging.DEBUG
    logging.getLogger().setLevel(log_level)

    dbconfig = load_config(db)
    con = psy.connect(cursor_factory=psycopg2.extras.RealDictCursor, **dbconfig)

    config = process_options(load_config(config))

    logging.info("Preprocessing...")
    preprocess(con, config)

    logging.info("Training...")
    deduper = train(con, config)

    logging.info("Creating blocking table...")
    create_blocking(deduper, con, config)

    logging.info("Clustering...")
    clustered_dupes = cluster(deduper, con, config)

    logging.info("Writing results...")
    write_results(clustered_dupes, con, config)

    logging.info("Applying results...")
    apply_results(con, config)

    # Close our database connection
    con.close()

    print('ran in', time.time() - START_TIME, 'seconds')


def load_config(filename):
    ext = os.path.splitext(filename)[1].lower()
    with open(filename) as f:
        if ext == '.json':
            return json.load(f)
        elif ext == '.yaml':
            return yaml.load(f)
        else:
            raise Exception('unknown filetype %s' % ext)


def process_options(c):
    config = dict()
    # Required fields
    for k in ('schema', 'table', 'key', 'fields'):
        if k not in c:
            raise Exception('Key ' + k + ' must be defined in the config file')
        config[k] = c[k]
    # Optional fields
    for k, default in (('interactions', []),
                       ('threshold', 0.5),
                       ('recall', 0.90),
                       ('merge_exact', []),
                       ('settings_file', 'dedup_postgres_settings'),
                       ('training_file', 'dedup_postgres_training.json'),
                       ('filter_condition', '1=1'),
                       ('num_cores', None)
                       ):
        config[k] = c.get(k, default)
    # Ensure that the merge_exact list is a list of lists
    if type(config['merge_exact']) is not list:
        raise Exception('merge_exact must be a list of columns')
    if len(config['merge_exact']) > 0 and type(config['merge_exact'][0]) is not list:
        config['merge_exact'] = [config['merge_exact']]
    # Add variable names to the field definitions, defaulting to the field
    for d in config['fields']:
        if 'variable name' not in d:
            d['variable name'] = d['field']
    # Add some handy computed values for convenience
    config['all_fields'] = config['fields'] + [
        {'type': 'Interaction', 'interaction variables': x} for x in config['interactions']]
    columns = set([x['field'] for x in config['fields']])
    config['columns'] = ', '.join(columns)
    config['all_columns'] = ', '.join(columns | set(['_unique_id']))
    return config


def preprocess(con, config):
    c = con.cursor()

    # Do an initial first pass and merge all exact duplicates
    c.execute("""CREATE SCHEMA IF NOT EXISTS {schema}""".format(**config))

    # TODO: Make the restriction configurable
    c.execute("""DROP TABLE IF EXISTS {schema}.entries_unique""".format(**config))
    c.execute("""CREATE TABLE {schema}.entries_unique AS (
                    SELECT {columns}, array_agg({key}) as src_ids FROM {table}
                    WHERE ({filter_condition})
                    GROUP BY {columns})""".format(**config))
    c.execute("ALTER TABLE {schema}.entries_unique "
              " ADD COLUMN _unique_id SERIAL PRIMARY KEY".format(**config))
    con.commit()


# Training
def train(con, config):
    if False:  # os.path.exists(settings_file):
        print('reading from ', config['settings_file'])
        with open(config['settings_file'], 'rb') as sf:
            return dedupe.StaticDedupe(sf, num_cores=2)
    # Create a new deduper object and pass our data model to it.
    deduper = dedupe.Dedupe(config['all_fields'], num_cores=config['num_cores'])

    # Named cursor runs server side with psycopg2
    cur = con.cursor('individual_select')

    cur.execute("SELECT {all_columns} FROM {schema}.entries_unique".format(**config))
    temp_d = dict((i, row) for i, row in enumerate(cur))

    deduper.sample(temp_d, 75000)
    del temp_d
    # If we have training data saved from a previous run of dedupe,
    # look for it an load it in.
    #
    # __Note:__ if you want to train from
    # scratch, delete the training_file
    if os.path.exists(config['training_file']):
        print('reading labeled examples from ', config['training_file'])
        with open(config['training_file']) as tf:
            deduper.readTraining(tf)

    # ## Active learning
    print('starting active labeling...')
    # Starts the training loop. Dedupe will find the next pair of records
    # it is least certain about and ask you to label them as duplicates
    # or not.

    # use 'y', 'n' and 'u' keys to flag duplicates
    # press 'f' when you are finished
    dedupe.convenience.consoleLabel(deduper)
    # When finished, save our labeled, training pairs to disk
    with open(config['training_file'], 'w') as tf:
        deduper.writeTraining(tf)

    # `recall` is the proportion of true dupes pairs that the learned
    # rules must cover. You may want to reduce this if your are making
    # too many blocks and too many comparisons.
    deduper.train(recall=config['recall'])

    with open(config['settings_file'], 'wb') as sf:
        deduper.writeSettings(sf)

    # We can now remove some of the memory hobbing objects we used
    # for training
    deduper.cleanupTraining()
    return deduper


# Blocking
def create_blocking(deduper, con, config):
    c = con.cursor()

    # To run blocking on such a large set of data, we create a separate table
    # that contains blocking keys and record ids
    print('creating blocking_map database')
    c.execute("DROP TABLE IF EXISTS {schema}.blocking_map".format(**config))
    c.execute("CREATE TABLE {schema}.blocking_map "
              "(block_key VARCHAR(200), _unique_id INT)".format(**config))  # TODO: THIS INT...
    # ... needs to be dependent upon the column type of entry_id

    # If dedupe learned a Index Predicate, we have to take a pass
    # through the data and create indices.
    print('creating inverted index')

    for field in deduper.blocker.index_fields:
        c2 = con.cursor('c2')
        c2.execute("SELECT DISTINCT {0} FROM {schema}.entries_unique".format(field, **config))
        field_data = (row[field] for row in c2)
        deduper.blocker.index(field_data, field)
        c2.close()

    # Now we are ready to write our blocking map table by creating a
    # generator that yields unique `(block_key, donor_id)` tuples.
    print('writing blocking map')

    c3 = con.cursor('donor_select2')
    c3.execute("SELECT {all_columns} FROM {schema}.entries_unique".format(**config))
    full_data = ((row['_unique_id'], row) for row in c3)
    b_data = deduper.blocker(full_data)

    # Write out blocking map to CSV so we can quickly load in with
    # Postgres COPY
    csv_file = tempfile.NamedTemporaryFile(prefix='blocks_', delete=False, mode='w')
    csv_writer = csv.writer(csv_file)
    csv_writer.writerows(b_data)
    c3.close()
    csv_file.close()

    f = open(csv_file.name, 'r')
    c.copy_expert("COPY {schema}.blocking_map FROM STDIN CSV".format(**config), f)
    f.close()

    os.remove(csv_file.name)

    con.commit()

    # Remove blocks that contain only one record, sort by block key and
    # donor, key and index blocking map.
    #
    # These steps, particularly the sorting will let us quickly create
    # blocks of data for comparison
    print('prepare blocking table. this will probably take a while ...')

    logging.info("indexing block_key")
    c.execute("CREATE INDEX blocking_map_key_idx "
              " ON {schema}.blocking_map (block_key)".format(**config))

    c.execute("DROP TABLE IF EXISTS {schema}.plural_key".format(**config))
    c.execute("DROP TABLE IF EXISTS {schema}.plural_block".format(**config))
    c.execute("DROP TABLE IF EXISTS {schema}.covered_blocks".format(**config))
    c.execute("DROP TABLE IF EXISTS {schema}.smaller_coverage".format(**config))

    # Many block_keys will only form blocks that contain a single
    # record. Since there are no comparisons possible withing such a
    # singleton block we can ignore them.
    logging.info("calculating {schema}.plural_key".format(**config))
    c.execute("CREATE TABLE {schema}.plural_key "
              "(block_key VARCHAR(200), "
              " block_id SERIAL PRIMARY KEY)".format(**config))

    c.execute("INSERT INTO {schema}.plural_key (block_key) "
              "SELECT block_key FROM {schema}.blocking_map "
              "GROUP BY block_key HAVING COUNT(*) > 1".format(**config))

    logging.info("creating {schema}.block_key index".format(**config))
    c.execute("CREATE UNIQUE INDEX block_key_idx "
              " ON {schema}.plural_key (block_key)".format(**config))

    logging.info("calculating {schema}.plural_block".format(**config))
    c.execute("CREATE TABLE {schema}.plural_block "
              "AS (SELECT block_id, _unique_id "
              " FROM {schema}.blocking_map INNER JOIN {schema}.plural_key "
              " USING (block_key))".format(**config))

    logging.info("adding _unique_id index and sorting index")
    c.execute("CREATE INDEX plural_block_id_idx "
              " ON {schema}.plural_block (_unique_id)".format(**config))
    c.execute("CREATE UNIQUE INDEX plural_block_block_id_id_uniq "
              " ON {schema}.plural_block (block_id, _unique_id)".format(**config))

    # To use Kolb, et.al's Redundant Free Comparison scheme, we need to
    # keep track of all the block_ids that are associated with a
    # particular donor records. We'll use PostgreSQL's string_agg function to
    # do this. This function will truncate very long lists of associated
    # ids, so we'll also increase the maximum string length to try to
    # avoid this
    # c.execute("SET group_concat_max_len = 4096")

    logging.info("creating {schema}.covered_blocks".format(**config))
    c.execute("CREATE TABLE {schema}.covered_blocks "
              " AS (SELECT _unique_id, "
              " string_agg(CAST(block_id AS TEXT), ',' ORDER BY block_id) "
              "   AS sorted_ids "
              " FROM {schema}.plural_block "
              " GROUP BY _unique_id)".format(**config))

    c.execute("CREATE UNIQUE INDEX covered_blocks_id_idx "
              "ON {schema}.covered_blocks (_unique_id)".format(**config))

    con.commit()

    # In particular, for every block of records, we need to keep
    # track of a donor records's associated block_ids that are SMALLER than
    # the current block's _unique_id. Because we ordered the ids when we did the
    # GROUP_CONCAT we can achieve this by using some string hacks.
    logging.info("creating {schema}.smaller_coverage".format(**config))
    c.execute("CREATE TABLE {schema}.smaller_coverage "
              " AS (SELECT _unique_id, block_id, "
              " TRIM(',' FROM split_part(sorted_ids, CAST(block_id AS TEXT), 1)) "
              "      AS smaller_ids "
              " FROM {schema}.plural_block INNER JOIN {schema}.covered_blocks "
              " USING (_unique_id))".format(**config))

    con.commit()


# Clustering
def candidates_gen(result_set):
    lset = set

    block_id = None
    records = []
    i = 0
    for row in result_set:
        if row['block_id'] != block_id:
            if records:
                yield records

            block_id = row['block_id']
            records = []
            i += 1

            if i % 10000 == 0:
                print(i, "blocks")
                print(time.time() - START_TIME, "seconds")

        smaller_ids = row['smaller_ids']

        if smaller_ids:
            smaller_ids = lset(smaller_ids.split(','))
        else:
            smaller_ids = lset([])

        records.append((row['_unique_id'], row, smaller_ids))

    if records:
        yield records


def cluster(deduper, con, config):
    c4 = con.cursor('c4')
    c4.execute("SELECT {all_columns}, block_id, smaller_ids FROM {schema}.smaller_coverage "
               "INNER JOIN {schema}.entries_unique "
               "USING (_unique_id) "
               "ORDER BY (block_id)".format(**config))

    return deduper.matchBlocks(candidates_gen(c4), threshold=config['threshold'])


# Writing out results
def write_results(clustered_dupes, con, config):
    c = con.cursor()
    # We now have a sequence of tuples of donor ids that dedupe believes
    # all refer to the same entity. We write this out onto an entity map
    # table
    c.execute("DROP TABLE IF EXISTS {schema}.entity_map".format(**config))

    c.execute("CREATE TABLE {schema}.entity_map "
              "(_unique_id INT, canon_id INT, "  # TODO: THESE INTS MUST BE DYNAMIC
              " cluster_score FLOAT, PRIMARY KEY(_unique_id))".format(**config))

    csv_file = tempfile.NamedTemporaryFile(prefix='entity_map_', delete=False,
                                           mode='w')
    csv_writer = csv.writer(csv_file)

    for cluster, scores in clustered_dupes:
        cluster_id = cluster[0]
        for donor_id, score in zip(cluster, scores):
            csv_writer.writerow([donor_id, cluster_id, score])

    csv_file.close()

    f = open(csv_file.name, 'r')
    c.copy_expert("COPY {schema}.entity_map FROM STDIN CSV".format(**config), f)
    f.close()

    os.remove(csv_file.name)

    con.commit()

    c.execute("CREATE INDEX head_index ON {schema}.entity_map (canon_id)".format(**config))
    con.commit()

    # Print out the number of duplicates found
    print('# duplicate sets')
    print(len(clustered_dupes))


# ## Payoff
def apply_results(con, config):
    c = con.cursor()
    # Dedupe only cares about matched records; it doesn't have a canonical id
    # for singleton records. So we create a mapping table between *all*
    # _unique_ids to their canonical_id (or back to themselves if singletons).
    c.execute("DROP TABLE IF EXISTS {schema}.map".format(**config))
    c.execute("CREATE TABLE {schema}.map "
              "AS SELECT COALESCE(canon_id, _unique_id) AS canon_id,"
              "_unique_id, "
              "COALESCE(cluster_score, 1.0) AS cluster_score "
              "FROM {schema}.entity_map "
              "RIGHT JOIN {schema}.entries_unique USING(_unique_id)".format(**config))

    # Merge clusters based upon exact matches of a subset of fields
    for cols in config['merge_exact']:
        exact_matches.merge('{}.map'.format(config['schema']),
                            '{}.entries_unique'.format(config['schema']),
                            cols,
                            con)

    # Add that integer id back to the unique_entries table
    c.execute(
        "ALTER TABLE {schema}.entries_unique DROP COLUMN IF EXISTS dedupe_id".format(**config))
    c.execute("ALTER TABLE {schema}.entries_unique ADD COLUMN dedupe_id INTEGER".format(**config))
    c.execute("UPDATE {schema}.entries_unique u SET dedupe_id = m.canon_id "
              "FROM {schema}.map m WHERE u._unique_id = m._unique_id".format(**config))
    con.commit()

    # And now map it the whole way back to the entries table
    # create a mapping between the unique entries and the original entries
    c.execute("DROP TABLE IF EXISTS {schema}.unique_map".format(**config))
    c.execute("CREATE TABLE {schema}.unique_map AS ( "
              "SELECT dedupe_id, unnest(src_ids) as {key} "
              "FROM {schema}.entries_unique)".format(**config))

    con.commit()

    c.execute("ALTER TABLE {table} DROP COLUMN IF EXISTS dedupe_id".format(**config))
    c.execute("ALTER TABLE {table} ADD COLUMN dedupe_id INTEGER".format(**config))
    c.execute("UPDATE {table} u SET dedupe_id = m.dedupe_id "
              "FROM {schema}.unique_map m WHERE u.{key} = m.{key}".format(**config))

    con.commit()
    c.close()


if __name__ == '__main__':
    main()



import tempfile

import pandas as pd
import numpy as np


def follow(id1, edges, visited=None, weak=True):
    if visited is None:
        visited = set()
    visited.add(id1)

    for row in edges[edges['id1'] == id1].values:
        if(row[1] not in visited):
            follow(row[1], edges, visited)

    if weak:
        for row in edges[edges['id2'] == id1].values:
            if(row[0] not in visited):
                follow(row[0], edges, visited)

    return visited


# if sparse (a lot of disconnected vertices) find those separately (faster)
def get_components(edges, vertices=None):
    if vertices is None:
        vertices = pd.DataFrame({'id': pd.concat((edges['id1'], edges['id2'])).unique()})

    visited = set()
    components = {}

    for id1 in vertices.values[:, 0]:
        if id1 not in visited:
            c = follow(id1, edges)
            visited.update(c)
            components[id1] = c

    return components


def components_dict_to_df(components):
    deduped = np.empty((0, 2), dtype=int)

    for id1 in components:
        deduped = np.append(deduped, [[id1, id2] for id2 in components[id1]], axis=0)

    deduped = pd.DataFrame(deduped, columns=['id1', 'id2'])
    return deduped


def merge(mapping_table, entries_table, exact_columns, con):
    edges = pd.read_sql("""
    with subset as (
        SELECT _unique_id, canon_id, {cols}
        FROM {entries} LEFT JOIN {mapping} using (_unique_id)
    )

    SELECT t1.canon_id id1, id2 from
    subset t1 JOIN
    (SELECT min(canon_id) id2, {cols} from subset group by {cols} having count(*) > 1) t
    using ({cols})
    where t1.canon_id > id2
    group by 1,2
    """.format(cols=', '.join(exact_columns), entries=entries_table, mapping=mapping_table), con)

    components = components_dict_to_df(get_components(edges))

    c = con.cursor()
    with tempfile.TemporaryFile() as f:
        components.to_csv(f, index=False, header=False)
        t = "merged_" + "_".join(exact_columns)
        c.execute("DROP TABLE IF EXISTS dedupe.{}".format(t))  # TODO SCHEMA NAME
        c.execute("CREATE TABLE dedupe.{} (id1 INT, id2 INT)".format(t))  # TODO DATA TYPE
        f.seek(0)
        c.copy_from(f, 'dedupe.' + t, sep=',')
    c.execute("""UPDATE {m} m SET
                     canon_id = t.id1
                 FROM dedupe.{t} t
                 WHERE m.canon_id = t.id2""".format(m=mapping_table, t=t))
    con.commit()



import yaml
import os
from os import path, system

# We'll shell out to `psql`, so set the environment variables for it:
with open("database.yml") as f:
    for k,v in yaml.load(f).items():
        os.environ['PG' + k.upper()] = v if v else ""
# And create the table from the csv file with psql
system("""psql -c "CREATE SCHEMA IF NOT EXISTS dedupe;" """)
system("""psql -c "DROP TABLE IF EXISTS dedupe.entries;" """)
# system("""csvsql --no-constraints -i postgresql --table entries --db-schema dedupe < test/people.csv | psql """)
system("""psql -c "
     CREATE TABLE "dedupe.entries" (
         uuid UUID,
         first_name VARCHAR,
         last_name VARCHAR,
         ssn VARCHAR(11),
         sex VARCHAR(1),
         dob VARCHAR(10),
         race VARCHAR,
         ethnicity VARCHAR
     );" """)
system("""psql -c "\copy dedupe.entries FROM 'test/people.csv' WITH CSV HEADER;" """)
system("""psql -c "alter table dedupe.entries add column entry_id SERIAL PRIMARY KEY;" """)
system("""psql -c "alter table dedupe.entries add column full_name VARCHAR;" """)
system("""psql -c "update dedupe.entries set full_name = first_name || ' ' || last_name;" """)
